#! /usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright(C) 2013 name <email>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio
from gi.repository import Unity
import gettext
import urllib.parse
import dbus
import hashlib
import unicodedata
import os
import shutil
import sqlite3
import dbus

APP_NAME = 'unity-scope-guayadeque'
LOCAL_PATH = '/usr/share/locale/'
gettext.bindtextdomain(APP_NAME, LOCAL_PATH)
gettext.textdomain(APP_NAME)
_ = gettext.gettext

GROUP_NAME = 'com.canonical.Unity.Scope.Music.Guayadeque'
UNIQUE_PATH = '/com/canonical/unity/scope/music/guayadeque'

SEARCH_HINT = _('Search Guayadeque')
NO_RESULTS_HINT = _('Sorry, there are no Guayadeque results that match your search.')
PROVIDER_CREDITS = _('')
SVG_DIR = '/usr/share/icons/unity-icon-theme/places/svg/'
PROVIDER_ICON = SVG_DIR + 'service-yelp.svg'
DEFAULT_RESULT_ICON = SVG_DIR + 'result-guayadeque.svg'
DEFAULT_RESULT_MIMETYPE = 'x-scheme-handler/man'
DEFAULT_RESULT_TYPE = Unity.ResultType.DEFAULT
GUAYADEQUE_DBFILE = os.getenv("HOME") + "/.guayadeque/Collections"

c1 = {'id': 'songs',
      'name': _('Songs'),
      'icon': SVG_DIR + 'group-installed.svg',
      'renderer': Unity.CategoryRenderer.VERTICAL_TILE}
c2 = {'id': 'albums',
      'name': _('Albums'),
      'icon': SVG_DIR + 'group-installed.svg',
      'renderer': Unity.CategoryRenderer.VERTICAL_TILE}
CATEGORIES = [c1, c2]

FILTERS = []

EXTRA_METADATA = []

REFRESH_TIMEOUT = 300
PREVIEW_PLAYER_DBUS_NAME = "com.canonical.Unity.Lens.Music.PreviewPlayer"
PREVIEW_PLAYER_DBUS_PATH = "/com/canonical/Unity/Lens/Music/PreviewPlayer"
PREVIEW_PLAYER_DBUS_IFACE = PREVIEW_PLAYER_DBUS_NAME

tracks = []


def get_music_from_guayadeque():
    """Parses Guayadeque's collections into a form we can use"""
    tracks = []
    # Copy guayadeque's database to a backup so we can run searches on that rather than the main database
    for collection in os.listdir(GUAYADEQUE_DBFILE):
        collection_tracks = []
        guayadeque_dbfile = GUAYADEQUE_DBFILE + "/" + collection + "/guayadeque.db"
        if not os.path.exists(guayadeque_dbfile):
            continue
        guayadeque_backupfile = GUAYADEQUE_DBFILE + "/" + collection + "/guayadeque-scope-backup.db"

        shutil.copy2(guayadeque_dbfile, guayadeque_backupfile)

        # Grab all the data we need from the backup
        conn = sqlite3.connect(guayadeque_backupfile)
        cursor = conn.cursor()
        # Go through the safe and grab track names, their uris, the artist name, the album title and the track's mimetypes
        # We'll have to call them all mp3s as the mimetype isn't explicitly in the database, but as long as it's an audio mimetype, it shouldn't matter
        cursor.execute('''SELECT songs.song_name, songs.song_path, songs.song_filename, songs.song_artist, songs.song_album, "taglib/mp3", songs.song_albumartist, covers.cover_path, songs.song_year, songs.song_genre, songs.song_number, songs.song_length
                          FROM songs, covers
                          WHERE songs.song_coverid = covers.cover_id
                          ORDER BY song_number''')
        collection_tracks = cursor.fetchall()
        cursor.close()
        for track in collection_tracks:
            tracks.append(track)
    print("Updated tracks from Guayadeque database")
    return tracks


def get_album_art(track):
    """Finds album art for the selected album
    """
    if not track[7] is None:
        return track[7]

    # If that fails, thumbnail any embedded album art and use that
    hashname = unicodedata.normalize("NFKD", unicode(track[5])) + "\t" + unicodedata.normalize("NFKD", unicode(track[3]))
    file_hash = hashlib.md5(hashname).hexdigest()
    tb_filename = os.path.join(os.path.expanduser("~/.cache/media-art"), ("album-" + file_hash)) + ".jpg"
    if os.path.exists(tb_filename):
        return tb_filename
    else:
        try:
            from mutagen import File
            audio = File(track[1])
            if "APIC:" in audio:
                artwork = audio.tags["APIC:"].data
                if not os.path.exists(os.path.expanduser("~/.cache/media-art")):
                    os.makedirs(os.path.expanduser("~/.cache/media-art"))
                with open(tb_filename, "wb") as img:
                    img.write(artwork)
                return tb_filename
            else:
                return "audio-x-generic"
        except:
            # Otherwise, return a generic audio icon
            return "audio-x-generic"


def search(search, filters):
    '''
    Search for help documents matching the search string
    '''
    results = []
    tracks = get_music_from_guayadeque()
    trackresults = []
    albumresults = []
    for track in tracks:
        title = u"" if track[0] is None else track[0]
        uri = u"" if track[1] is None else track[1] + track[2]
        try:
            uri = uri.decode('utf-8')
        except:
            pass
        artist = u"" if track[2] is None else track[3]
        album = u"" if track[3] is None else track[4]
        mimetype = u"" if track[4] is None else track[5]
        albumartist = u"" if track[5] is None else track[6]
        trackname = title + u" - " + album + u" - " + artist
        if search.lower() in trackname.lower():
            try:
                albumart = get_album_art(track)
            except:
                albumart = ""
            albumuri = "album://" + albumartist + "/" + album
            if track not in trackresults:
                results.append({'uri': uri,
                                'icon': albumart,
                                'category': 0,
                                'mimetype': mimetype,
                                'title': title,
                                'comment': artist})
                trackresults.append(track)

            if album not in albumresults:
                results.append({'uri': albumuri,
                                'icon': albumart,
                                'category': 1,
                                'mimetype': mimetype,
                                'title': album,
                                'comment': artist})
                albumresults.append(album)
    return results


def activate(scope, uri):
    import subprocess
    albumtracks = []
    albumtracks.append("guayadeque")
    # If uri starts with album:// then we need to play all the songs on it
    if not uri.startswith("album://") == -1:
        for track in tracks:
            album = "album://" + track[6] + "/" + track[4]
            if not album.find(uri) == -1:
                albumtrack = urllib.unquote(str(track[1]) + str(track[2]))
                albumtracks.append(albumtrack)
        subprocess.Popen(albumtracks)
    else:
        albumtracks.append(uri)
        subprocess.Popen(albumtracks)
    try:
        session_bus = dbus.SessionBus()
        player = session_bus.get_object('org.mpris.guayadeque', '/Player')
        iface = dbus.Interface(player, dbus_interface='org.freedesktop.MediaPlayer')
        iface.Play()
    except:
        pass

    return Unity.ActivationResponse(handled=Unity.HandledType.HIDE_DASH, goto_uri='')


def show_in_folder(scope, uri):
    """ Shows the folder containing the selected track as requested from the Preview
    """
    if uri.startswith("album://"):
        for track in tracks:
            album = "album://" + track[2] + "/" + track[3]
            if not album.find(uri) == -1:
                filename = track[1]
                continue
    else:
        filename = uri
    dirname = os.path.dirname(filename)
    dirname = dirname.replace("%20", "\ ")
    os.system("xdg-open '%s'" % str(dirname))
    return Unity.ActivationResponse(handled=Unity.HandledType.HIDE_DASH, goto_uri='')


def preview_uri(scope, uri):
    """Preview request handler"""
    albumtracks = []
    isalbum = False
    if uri.startswith("album://"):
        isalbum = True
        for track in tracks:
            album = "album://" + track[2] + "/" + track[3]
            if not album.find(uri) == -1:
                albumtracks.append(track)
        albumtracks.sort(key=lambda track: int(track[7]))
    else:
        for track in tracks:
            album = "file://" + track[1]
            if not album.find(uri) == -1:
                albumtracks.append(track)
    iteration = model.get_first_iter()
    end_iter = model.get_last_iter()
    while iteration != end_iter:
        if model.get_value(iteration, 0) == uri:
            title = model.get_value(iteration, 5)
            description = model.get_value(iteration, 6)
            if model.get_value(iteration, 1) == "musique":
                image = "file:///usr/share/icons/hicolor/scalable/apps/audacious.svg"
            else:
                image = "file://%s" % model.get_value(iteration, 1)

            preview = Unity.MusicPreview.new(title, description, None)
            preview.props.image_source_uri = image
            for albumtrack in albumtracks:
                if isalbum:
                    track = Unity.TrackMetadata.full("file://" + urllib.parse.unquote(str(albumtrack[1])),  # uri
                                                     int(albumtrack[7]),                              # track number
                                                     albumtrack[0],                                   # track title
                                                     albumtrack[2],                                   # artist
                                                     albumtrack[3],                                   # album
                                                     int(albumtrack[8]) / 1000)                       # track length
                else:
                    preview = Unity.MusicPreview.new(albumtrack[0], "", None)
                    preview.props.image_source_uri = image
                    track = Unity.TrackMetadata.full("file://" + urllib.parse.unquote(str(albumtrack[1])),
                                                     int(albumtrack[7]),
                                                     albumtrack[0],
                                                     albumtrack[2],
                                                     albumtrack[3],
                                                     int(albumtrack[8]) / 1000)
                preview.add_track(track)

            # Add the "Play" action
            play_action = Unity.PreviewAction.new("activate_uri", "Play", None)
            play_action.connect("activated", activate)
            preview.add_action(play_action)

            # Add the "Show in folder" action
            show_action = Unity.PreviewAction.new("show_in_folder", "Show In Folder", None)
            show_action.connect("activated", show_in_folder)
            preview.add_action(show_action)

            preview.connect("play", play)
            preview.connect("pause", pause)
            preview.connect("closed", closed)
            break
        iteration = model.next(iteration)
    if preview is None:
        print("Couldn't find model row for requested preview uri: '%s'", uri)
    return preview


def play(preview, uri):
    """Plays the selected track as selected in the Preview"""
    player = self.bus.get_object(PREVIEW_PLAYER_DBUS_NAME, PREVIEW_PLAYER_DBUS_PATH)
    dbus.Interface(player, PREVIEW_PLAYER_DBUS_IFACE).Play(uri)


def pause(preview, uri):
    """Pauses the selected track as selected in the Preview"""
    player = self.bus.get_object(PREVIEW_PLAYER_DBUS_NAME, PREVIEW_PLAYER_DBUS_PATH)
    dbus.Interface(player, PREVIEW_PLAYER_DBUS_IFACE).Pause()


def closed(preview):
    """Stops playing when the previre is closed"""
    player = self.bus.get_object(PREVIEW_PLAYER_DBUS_NAME, PREVIEW_PLAYER_DBUS_PATH)
    dbus.Interface(player, PREVIEW_PLAYER_DBUS_IFACE).Close()

# Classes below this point establish communication
# with Unity, you probably shouldn't modify them.


class MySearch(Unity.ScopeSearchBase):
    def __init__(self, search_context):
        super(MySearch, self).__init__()
        self.set_search_context(search_context)

    def do_run(self):
        '''
        Adds results to the model
        '''
        try:
            result_set = self.search_context.result_set
            for i in search(self.search_context.search_query,
                            self.search_context.filter_state):
                if not 'uri' in i or not i['uri'] or i['uri'] == '':
                    continue
                if not 'icon' in i or not i['icon'] or i['icon'] == '':
                    i['icon'] = DEFAULT_RESULT_ICON
                if not 'mimetype' in i or not i['mimetype'] or i['mimetype'] == '':
                    i['mimetype'] = DEFAULT_RESULT_MIMETYPE
                if not 'result_type' in i or not i['result_type'] or i['result_type'] == '':
                    i['result_type'] = DEFAULT_RESULT_TYPE
                if not 'category' in i or not i['category'] or i['category'] == '':
                    i['category'] = 0
                if not 'title' in i or not i['title']:
                    i['title'] = ''
                if not 'comment' in i or not i['comment']:
                    i['comment'] = ''
                if not 'dnd_uri' in i or not i['dnd_uri'] or i['dnd_uri'] == '':
                    i['dnd_uri'] = i['uri']
                i['metadata'] = {}
                if EXTRA_METADATA:
                    for e in i:
                        for m in EXTRA_METADATA:
                            if m['id'] == e:
                                i['metadata'][e] = i[e]
                i['metadata']['provider_credits'] = GLib.Variant('s', PROVIDER_CREDITS)
                result = Unity.ScopeResult.create(str(i['uri']), str(i['icon']),
                                                  i['category'], i['result_type'],
                                                  str(i['mimetype']), str(i['title']),
                                                  str(i['comment']), str(i['dnd_uri']),
                                                  i['metadata'])
                result_set.add_result(result)
        except Exception as error:
            print(error)


class Scope(Unity.AbstractScope):
    def __init__(self):
        Unity.AbstractScope.__init__(self)

    def do_get_search_hint(self):
        return SEARCH_HINT

    def do_get_schema(self):
        '''
        Adds specific metadata fields
        '''
        schema = Unity.Schema.new()
        if EXTRA_METADATA:
            for m in EXTRA_METADATA:
                schema.add_field(m['id'], m['type'], m['field'])
        #FIXME should be REQUIRED for credits
        schema.add_field('provider_credits', 's', Unity.SchemaFieldType.OPTIONAL)
        return schema

    def do_get_categories(self):
        '''
        Adds categories
        '''
        cs = Unity.CategorySet.new()
        if CATEGORIES:
            for c in CATEGORIES:
                cat = Unity.Category.new(c['id'], c['name'],
                                         Gio.ThemedIcon.new(c['icon']),
                                         c['renderer'])
                cs.add(cat)
        return cs

    def do_get_filters(self):
        '''
        Adds filters
        '''
        fs = Unity.FilterSet.new()
        #if FILTERS:
        #
        return fs

    def do_get_group_name(self):
        return GROUP_NAME

    def do_get_unique_name(self):
        return UNIQUE_PATH

    def do_create_search_for_query(self, search_context):
        se = MySearch(search_context)
        return se


def load_scope():
    return Scope()
